#include "MeshLoader.h"
#include "HashCustom.h"


const std::vector<Node>& MeshLoader::Get_Node() const {
    return nodes;
}

const std::vector<FiniteElement>& MeshLoader::Get_Finite_Elements() const {
    return finite_elements;
}

const std::vector<BoundaryFiniteElement>& MeshLoader::Get_Boundary() const {
    return boundary;
}

std::set<int> MeshLoader::Get_Nodes_By_Surface(int t_surface_id) const
{
	std::set<int> res;
	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();

	while (first_to_find != last_to_find) {
		auto found_boundary = std::find_if(first_to_find, last_to_find,
			[t_surface_id](const BoundaryFiniteElement& cur_boundary) {
				return cur_boundary.surface_id == t_surface_id;
			});

		first_to_find = found_boundary;
		if (found_boundary != last_to_find) {
			for (const auto& node_in_boundary : found_boundary->nodes_id) res.insert(node_in_boundary);
			++first_to_find;
		}
	}
	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_ID(int t_node_id1, int t_node_id2, int t_node_id3) const
{
	std::vector<int> res;
	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_node_id1, t_node_id2, t_node_id3](const FiniteElement& elem) {
				return std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id1) != elem.nodes_id.end()
					&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id2) != elem.nodes_id.end()
					&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id3) != elem.nodes_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_Edge(int t_node_id1, int t_node_id2) const
{
	std::vector<int> res;

	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_node_id1, t_node_id2](const FiniteElement& elem) {
				return std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id1) != elem.nodes_id.end()
					&& std::find(elem.nodes_id.begin(), elem.nodes_id.end(), t_node_id2) != elem.nodes_id.end();
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Elements_By_Material(int t_material_id) const
{
	std::vector<int> res;

	auto first_to_find = finite_elements.begin();
	auto last_to_find = finite_elements.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_material_id](const FiniteElement& elem) {
				return elem.material_id == t_material_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

std::vector<int> MeshLoader::Get_Boundary_By_Surface(int t_surface_id) const
{
	std::vector<int> res;

	auto first_to_find = boundary.begin();
	auto last_to_find = boundary.end();

	while (first_to_find != last_to_find) {
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[t_surface_id](const BoundaryFiniteElement& elem) {
				return elem.surface_id == t_surface_id;
			});

		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			res.push_back(found_elem->element_id);
			++first_to_find;
		}
	}

	return res;
}

void MeshLoader::Insert_Middle_Nodes()
{

	std::unordered_set <Edge> examined_edges;
	for (auto& elem : finite_elements) {
		std::vector<int> cur_nodes_id = elem.nodes_id;
		for (auto first_node = 0; first_node < 4; ++first_node)
			for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
				Edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);
				if (examined_edges.insert(cur_edge).second) {
					Node new_node = Get_Middle_Node(cur_edge);
					cur_edge.UpdateMiddleNode(new_node.node_id);
					nodes.push_back(new_node);
				}
				else {
					elem.nodes_id.push_back(cur_edge.middle_node);
				}
			}
	}

	for (auto& elem : boundary) {
		std::vector<int> cur_nodes_id = elem.nodes_id;
		for (auto first_node = 0; first_node < 3; ++first_node)
			for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
				auto found_edge = examined_edges.find({ cur_nodes_id[first_node], cur_nodes_id[second_node] });
				if (found_edge == examined_edges.end())
					found_edge = examined_edges.find({ cur_nodes_id[second_node], cur_nodes_id[first_node] });
				elem.nodes_id.push_back(found_edge->middle_node);
			}
	}
}

std::vector<std::set<int>> MeshLoader::Get_Adjacent_Nodes() const
{
	std::vector<std::set<int>> res(nodes.size() + 1);

	for (const auto& cur_elem : finite_elements) {
		for (auto cur_node_id : cur_elem.nodes_id)
			for (auto adj_node_id : cur_elem.nodes_id)
				if (cur_node_id != adj_node_id)
					res[cur_node_id].insert(adj_node_id);
	}
	return res;
}


std::array<double, 3> MeshLoader::Get_Middle_Of_Edge(const Edge& t_edge)
{
	return
	{ (get<0>(nodes[t_edge.nodes.first - 1].coors) + get<0>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<1>(nodes[t_edge.nodes.first - 1].coors) + get<1>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2
	, (get<2>(nodes[t_edge.nodes.first - 1].coors) + get<2>(nodes[t_edge.nodes.second - 1].coors)) * 1 / 2 };
}

Node MeshLoader::Get_Middle_Node(const Edge& cur_edge)
{
	Node new_node(nodes.size() + 1, Get_Middle_Of_Edge(cur_edge), 0);
	return new_node;
}

Node& MeshLoader::Get_Node_By_ID(int t_id)
{
	return nodes.at(t_id);
}

std::array<double, 3> MeshLoader::Get_Coors_Node_By_ID(int t_id)
{
	return nodes.at(t_id).coors;
}



FiniteElement MeshLoader::Get_Elements_By_Min_Ratio()
{
	auto res = std::ranges::max(finite_elements,
		std::less<>{},
		[this](const FiniteElement& t_elem) { return Get_Min_Height_In_Element(t_elem) / Get_Max_Height_In_Element(t_elem); });
	return res;
}

template <typename _T>
double get_area(const _T& t_range) {
	auto x = t_range.at(0);
	auto y = t_range.at(1);
	auto z = t_range.at(2);
	double a = sqrt(pow(x[1] - y[1], 2) + pow(x[2] - y[2], 2) + pow(x[0] - y[0], 2));
	double b = sqrt(pow(y[1] - z[1], 2) + pow(y[2] - z[2], 2) + pow(y[0] - z[0], 2));
	double c = sqrt(pow(z[1] - x[1], 2) + pow(z[2] - x[2], 2) + pow(z[0] - x[0], 2));

	double p = (a + b + c) / 2;
	return sqrt(p * (p - a) * (p - b) * (p - c));
}

double MeshLoader::Get_Max_Height_In_Element(const FiniteElement& t_elem) const
{
	double max_base = 0;
	for (auto cur_node : t_elem.nodes_id) {
		auto _nodes_in_base = t_elem.nodes_id
			| std::ranges::views::filter([cur_node](auto cur_idx) { return cur_idx != cur_node; })
			| std::ranges::views::transform([this](int _idx) { return nodes[_idx - 1].coors; });
		std::vector<std::array<double, 3>> nodes_in_base;
		std::ranges::copy(_nodes_in_base, std::back_inserter(nodes_in_base));
		auto cur_base = get_area(nodes_in_base);
		if (cur_base > max_base) max_base = cur_base;
	}
	return max_base;
}

template <typename _T>
double Get_Height(const std::array<double, 3>& p, const _T& p_range) {
	auto x = p_range.at(0);
	auto y = p_range.at(1);
	auto z = p_range.at(2);
	double A = (x[1] - y[1]) * (x[2] - z[2]) - (x[1] - z[1]) * (x[2] - y[2]);
	double B = -((x[0] - y[0]) * (x[2] - z[2]) - (x[0] - z[0]) * (x[2] - y[2]));
	double C = (x[0] - y[0]) * (x[1] - z[1]) - (x[0] - z[0]) * (x[1] - y[1]);
	double D = -x[0] * A + x[1] * B - x[2] * C;
	return fabs((A * (p[0] - x[0]) + B * (p[1] - x[1]) + C * (p[2] - x[2]) + D) / sqrt(A * A + B * B + C * C));
}

double MeshLoader::Get_Min_Height_In_Element(const FiniteElement& t_elem) const
{
	double max_height = 0;
	for (auto cur_node : t_elem.nodes_id) {
		auto _nodes_in_base = t_elem.nodes_id
			| std::ranges::views::filter([cur_node](auto cur_idx) { return cur_idx != cur_node; })
			| std::ranges::views::transform([this](int _idx) { return nodes[_idx - 1].coors; });
		std::vector<std::array<double, 3>> nodes_in_base;
		std::ranges::copy(_nodes_in_base, std::back_inserter(nodes_in_base));
		double cur_height = Get_Height(nodes[cur_node - 1].coors, nodes_in_base);
		if (max_height < cur_height) max_height = cur_height;
	}
	return max_height;
}
